import logging

from vortex.TupleAction import TupleActionABC

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField

logger = logging.getLogger(__name__)


@addTupleType
class InternalUserUpdatePasswordAction(TupleActionABC):
    __tupleType__ = userPluginTuplePrefix + "InternalUserUpdatePasswordAction"

    userId: int = TupleField()
    newPassword: str = TupleField()



from typing import List

from vortex.TupleAction import TupleActionABC

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField


@addTupleType
class UserLogoutAction(TupleActionABC):
    __tupleType__ = userPluginTuplePrefix + "UserLogoutAction"
    userName: str = TupleField()
    deviceToken: str = TupleField()

    isFieldService: bool = TupleField()
    isOfficeService: bool = TupleField()

    #: A list of accepted warning keys
    # If any server side warnings occur and they are in this list then the logoff
    # continues.
    acceptedWarningKeys: List[str] = TupleField()



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

'''
import logging

from sqlalchemy import Column
from sqlalchemy import Integer, String
from sqlalchemy.sql.schema import Index
from vortex.Tuple import Tuple, addTupleType, TupleField

from sqlalchemy import orm

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class DirectoryTuple(DeclarativeBase, Tuple):
    """ Directory Table

    This table contains different directories that can be authenticated against.

    """
    __tupleType__ = userPluginTuplePrefix + 'DirectoryTuple'
    __tablename__ = 'Directory'

    TYPE_INTERNAL = 1
    # TYPE_ACTIVE_DIRECYORY = 2 // TODO

    id = Column(Integer, primary_key=True, autoincrement=True)
    key = Column(String(30), unique=True, nullable=False)
    name = Column(String(30), unique=True, nullable=False)
    type = Column(String, nullable=False)

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)

    # __table_args__ = (
    #     Index("idx_UserDbTable_unique_index", userName,
    #           unique=True),
    # )
'''



import logging

from peek_core_user._private.PluginNames import (
    userPluginFilt,
    userPluginActionProcessorName,
)
from peek_core_user._private.server.controller.MainController import (
    MainController,
)
from vortex.handler.TupleActionProcessor import TupleActionProcessor

logger = logging.getLogger(__name__)


def makeTupleActionProcessorHandler(mainController: MainController):
    processor = TupleActionProcessor(
        tupleActionProcessorName=userPluginActionProcessorName,
        additionalFilt=userPluginFilt,
        defaultDelegate=mainController,
    )
    return processor



"""
{'objectClass': [b'top', b'person', b'organizationalPerson', b'user'], 'cn': [b'attest'],
 'givenName': [b'attest'],
 'distinguishedName': [b'CN=attest,OU=testou,DC=synad,DC=synerty,DC=com'],
 'instanceType': [b'4'], 'whenCreated': [b'20170505160836.0Z'],
 'whenChanged': [b'20190606130621.0Z'], 'displayName': [b'attest'],
 'uSNCreated': [b'16498'],
 'memberOf': [b'CN=Domain Admins,CN=Users,DC=synad,DC=synerty,DC=com',
              b'CN=Enterprise Admins,CN=Users,DC=synad,DC=synerty,DC=com',
              b'CN=Administrators,CN=Builtin,DC=synad,DC=synerty,DC=com'],
 'uSNChanged': [b'73784'], 'name': [b'attest'],
 'objectGUID': [b'\xee\x1bV\x8dQ\xackE\x82\xd9%_\x18\xadjO'],
 'userAccountControl': [b'66048'], 'badPwdCount': [b'0'], 'codePage': [b'0'],
 'countryCode': [b'0'], 'badPasswordTime': [b'132042996316396717'], 'lastLogoff': [b'0'],
 'lastLogon': [b'132042996806397639'], 'pwdLastSet': [b'132042997225927009'],
 'primaryGroupID': [b'513'], 'objectSid': [
    b'\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00D:3|X\x8f\xc7\x08\xe6\xeaV\xc8Q\x04\x00\x00'],
 'adminCount': [b'1'], 'accountExpires': [b'9223372036854775807'], 'logonCount': [b'36'],
 'sAMAccountName': [b'attest'], 'sAMAccountType': [b'805306368'],
 'userPrincipalName': [b'attest@synad.synerty.com'], 'lockoutTime': [b'0'],
 'objectCategory': [b'CN=Person,CN=Schema,CN=Configuration,DC=synad,DC=synerty,DC=com'],
 'dSCorePropagationData': [b'20190606130621.0Z', b'20190606130016.0Z',
                           b'20170506090346.0Z', b'16010101000000.0Z'],
 'lastLogonTimestamp': [b'132042996806397639']}
"""

import logging
from typing import List
from typing import Optional
from typing import Tuple

import ldap
from twisted.cred.error import LoginFailed

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage import LdapSetting
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.tuples.LdapLoggedInUserTuple import (
    LdapLoggedInUserTuple,
)
from peek_core_user.tuples.constants.UserAuthTargetEnum import (
    UserAuthTargetEnum,
)

logger = logging.getLogger(__name__)


def checkLdapAuth(
    username: str,
    password: str,
    ldapSetting: LdapSetting,
    userUuid: Optional[str],
) -> Tuple[List[str], LdapLoggedInUserTuple]:
    try:
        conn = ldap.initialize(ldapSetting.ldapUri)
        conn.protocol_version = 3
        conn.set_option(ldap.OPT_REFERRALS, 0)

        # make the connection
        conn.simple_bind_s(
            "%s@%s" % (username.split("@")[0], ldapSetting.ldapDomain), password
        )
        logger.info(
            "User=%s, Connected to LDAP server %s",
            username,
            ldapSetting.ldapDomain,
        )

        if userUuid:
            ldapFilter = (
                "(&(objectCategory=person)(objectClass=user)(objectSid=%s))"
                % userUuid
            )
        else:
            ldapFilter = (
                "(&(objectCategory=person)(objectClass=user)(sAMAccountName=%s))"
                % username.split("@")[0]
            )
        logger.debug("User=%s, LDAP user query: %s", username, ldapFilter)

        dcParts = ",".join(
            ["DC=%s" % part for part in ldapSetting.ldapDomain.split(".")]
        )

        ldapBases = []
        if ldapSetting.ldapOUFolders:
            ldapBases += _makeLdapBase(
                ldapSetting.ldapOUFolders, username, "OU"
            )
        if ldapSetting.ldapCNFolders:
            ldapBases += _makeLdapBase(
                ldapSetting.ldapCNFolders, username, "CN"
            )

        if not ldapBases:
            logger.debug(
                "User=%s, LDAP OU and/or CN search paths must be set.", username
            )
            raise LoginFailed(
                "LDAPAuth: LDAP OU and/or CN search paths must be set."
            )

        userDetails = None
        for ldapBase in ldapBases:
            ldapBase = "%s,%s" % (ldapBase, dcParts)
            logger.debug(
                "User=%s, Searching in LDAP Base: %s, for LDAP Filter: %s",
                username,
                ldapBase,
                ldapFilter,
            )

            try:
                # Example Base: 'CN=atuser1,CN=Users,DC=synad,DC=synerty,DC=com'
                userDetails = conn.search_st(
                    ldapBase, ldap.SCOPE_SUBTREE, ldapFilter, None, 0, 10
                )

                if userDetails:
                    break

                logger.debug(
                    "User=%s, Checking next, user was not found in: %s",
                    username,
                    ldapBase,
                )

            except ldap.NO_SUCH_OBJECT:
                logger.warning(
                    "User=%s, CN or OU doesn't exist : %s", username, ldapBase
                )

    except ldap.NO_SUCH_OBJECT:
        logger.info(
            "User=%s, was not found in any LDAP bases, NO_SUCH_OBJECT", username
        )
        raise LoginFailed(
            "LDAPAuth: A user with username %s was not found, ask admin to "
            "check Peek logs" % username
        )

    except ldap.INVALID_CREDENTIALS:
        logger.info(
            "User=%s, provided an incorrect username or password, INVALID_CREDENTIALS"
        )
        raise LoginFailed(
            "LDAPAuth: Username or password is incorrect for %s" % username
        )

    if not userDetails:
        logger.info(
            "User=%s, was not found in any LDAP bases, 'not userDetails'",
            username,
        )
        raise LoginFailed(
            "LDAPAuth: User %s doesn't belong to the correct CN/OUs" % username
        )

    userDetails = userDetails[0][1]

    distinguishedName = userDetails.get("distinguishedName")[0].decode()
    primaryGroupId = userDetails.get("primaryGroupID")[0].decode()
    objectSid = userDetails.get("objectSid")[0]
    # python-ldap doesn't include key `memberOf` in search result
    #  if the user doesn't belong to any groups.
    memberOfSet = set(userDetails.get("memberOf", []))

    decodedSid = _decodeSid(objectSid)
    primaryGroupSid = (
        "-".join(decodedSid.split("-")[:-1]) + "-" + primaryGroupId
    )

    ldapFilter = "(objectSid=%s)" % primaryGroupSid
    logger.debug(
        "User=%s, Primary group details LDAP filter: %s", username, ldapFilter
    )
    primGroupDetails = conn.search_st(
        dcParts, ldap.SCOPE_SUBTREE, ldapFilter, None, 0, 10
    )
    memberOfSet.add(primGroupDetails[0][1].get("distinguishedName")[0])

    # find all it's groups and groups of those groups
    # The magic number in this filter allows us to fetch the groups of
    # a group.
    ldapFilter = (
        "(&(objectCategory=group)(member:1.2.840.113556.1.4.1941:=%s))"
        % (_escapeParensForLdapFilter(distinguishedName),)
    )
    logger.debug(
        "User=%s, Using recursive groups filter: %s", username, ldapFilter
    )
    logger.info("Fetching groups from the LDAP server for user %s", username)
    groupDetails = conn.search_st(
        ",".join(distinguishedName.split(",")[1:]),
        ldap.SCOPE_SUBTREE,
        ldapFilter,
        None,
        0,
        10,
    )

    if groupDetails:
        for group in groupDetails:
            groupMemberOf = group[1].get("memberOf", [])
            memberOfSet.update(groupMemberOf)

    groups = []
    for memberOf in memberOfSet:
        group = memberOf.decode().split(",")[0]
        if "=" in group:
            group = group.split("=")[1]
        groups.append(group)

    logger.debug("User %s, is a member of groups: %s", username, groups)

    userTitle = None
    if userDetails.get("displayName"):
        userTitle = userDetails["displayName"][0].decode()

    email = None
    if userDetails.get("userPrincipalName"):
        email = userDetails["userPrincipalName"][0].decode()

    if not userUuid:
        userUuid = _decodeSid(objectSid)

    if ldapSetting.ldapGroups:
        ldapGroups = set([s.strip() for s in ldapSetting.ldapGroups.split(",")])

        logger.debug(
            "User=%s, Checking if user is a member of groups: %s",
            username,
            groups,
        )

        if not ldapGroups & set(groups):
            logger.info(
                "User=%s, is not a member of any authorised group, 'not ldapGroups & set(groups)'",
                username,
            )
            raise LoginFailed(
                "User %s is not a member of an authorised group" % username
            )

        logger.debug(
            "User=%s, is a member of specified groups. Proceeding with login",
            username,
        )

    ldapLoggedInUser = LdapLoggedInUserTuple(
        username=username,
        userTitle=userTitle,
        userUuid=userUuid,
        email=email,
        ldapName=ldapSetting.ldapTitle,
        objectSid=objectSid,
        ldapDomain=ldapSetting.ldapDomain,
    )

    return list(groups), ldapLoggedInUser


def _decodeSid(sid: [bytes]) -> str:
    strSid = "S-"
    sid = iter(sid)

    # Byte 0 is the revision
    revision = next(sid)
    strSid += "%s" % (revision,)

    # Byte 1 is the count of sub-authorities
    countSubAuths = next(sid)

    # Byte 2-7 (big endian) form the 48-bit authority code
    bytes27 = [next(sid) for _ in range(2, 8)]
    authority = int.from_bytes(bytes27, byteorder="big")
    strSid += "-%s" % (authority,)

    for _ in range(countSubAuths):
        # Each is 4 bytes (32-bits) in little endian
        subAuthBytes = [next(sid) for _ in range(4)]
        subAuth = int.from_bytes(subAuthBytes, byteorder="little")
        strSid += "-%s" % (subAuth,)

    return strSid


def _escapeParensForLdapFilter(value: str) -> str:
    """Escape parenthesis () in a string

    Escape special characters in a string to be able to use it as a value
    in an LDAP filter. `(` are replaced with \28 and `)` are replaced
    with \29 and so on.

    Reference: https://tools.ietf.org/search/rfc2254#page-5

    :return: Escaped string
    """
    # The \ character must always be escaped first
    value = value.replace("\\", "\\5C")

    value = value.replace("(", "\\28")
    value = value.replace(")", "\\29")
    value = value.replace("*", "\\2A")
    value = value.replace("\0", "\\00")
    return value


def _makeLdapBase(ldapFolders, userName, propertyName):
    try:
        ldapBases = []
        for folder in ldapFolders.split(","):
            folder = folder.strip()
            if not folder:
                continue

            parts = []
            for part in folder.split("/"):
                part = part.strip()
                if not part:
                    continue
                parts.append("%s=%s" % (propertyName, part))

            ldapBases.append(",".join(reversed(parts)))

        return ldapBases

    except Exception as e:
        logger.error(
            "Login failed for %s, failed to parse LDAP %s Folders setting",
            propertyName,
            userName,
        )

        logger.exception(e)

        raise LoginFailed(
            "An internal error occurred, ask admin to check Attune logs"
        )



import logging

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user.tuples.UserDetailTuple import UserDetailTuple
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

logger = logging.getLogger(__name__)


@addTupleType
class DeviceWithUserDetailsTuple(Tuple):
    """Field device tokens with logged-ion user details

    This tuple is sent to index peek-core-search for field crew/device
    related searches
    """

    __tupleType__ = userPluginTuplePrefix + "DeviceWithUserDetailsTuple"

    userDetails: UserDetailTuple = TupleField()
    deviceToken: str = TupleField()



"""Added User Import Source

Peek Plugin Database Migration Script

Revision ID: 43df0e05c728
Revises: b09fe7b3b31b
Create Date: 2022-03-10 12:38:29.301419

"""

# revision identifiers, used by Alembic.
from sqlalchemy.dialects import postgresql

revision = "43df0e05c728"
down_revision = "b09fe7b3b31b"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # add user authentication target
    op.add_column(
        "InternalUser",
        sa.Column(
            "authenticationTarget",
            sa.String(),
            nullable=True,
        ),
        schema="core_user",
    )

    # this migration does not work if there're existing users created by LDAP.
    #  Peek Admin needs to delete them before these users log in.
    op.execute(
        """
        UPDATE core_user."InternalUser"
        SET "authenticationTarget" = 'INTERNAL';
        """
    )

    # set `authenticationTarget` not null
    op.alter_column(
        "InternalUser",
        "authenticationTarget",
        nullable=False,
        schema="core_user",
    )

    # add user import source
    op.add_column(
        "InternalUser",
        sa.Column(
            "importSource",
            sa.String(),
            nullable=True,
        ),
        schema="core_user",
    )

    # START of labeling user import sources
    #  enmac-imported users
    op.execute(
        """
        UPDATE core_user."InternalUser"
        SET "importSource" = 'ENMAC'
        WHERE
            "importHash" LIKE 'pof_loader:%';
        """
    )
    # LDAP users
    op.execute(
        """
        UPDATE core_user."InternalUser"
        SET "importSource" = 'LDAP'
        WHERE
            "importHash" LIKE 'peek_core_user.LdapAuth:%';
        """
    )
    #  other users - undetermined
    op.execute(
        """
        UPDATE core_user."InternalUser"
        SET "importSource" = 'UNDETERMINED'
        WHERE
            "importSource" IS NULL;
        """
    )
    # END of labeling user import sources

    # set `importSource` not null
    op.alter_column(
        "InternalUser", "importSource", nullable=False, schema="core_user"
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("InternalUser", "authenticationTarget", "core_user")
    op.drop_column("InternalUser", "importSource", "core_user")
    # ### end Alembic commands ###



"""add otp-group

Peek Plugin Database Migration Script

Revision ID: 8fbb7b3fa4d8
Revises: 3cbb84cc31d5
Create Date: 2024-12-02 19:26:32.558867

"""

# revision identifiers, used by Alembic.
revision = "8fbb7b3fa4d8"
down_revision = "3cbb84cc31d5"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.execute(
        """
        INSERT INTO core_user."InternalGroup"( "groupName", "groupTitle")
        VALUES ('peek-field-app-otp-login',	'Use One Time Password')	
    """
    )


def downgrade():
    pass



import logging
from typing import List

from twisted.cred.error import LoginFailed
from twisted.internet.defer import inlineCallbacks

from peek_core_user._private.server.auth_connectors.InternalAuth import (
    InternalAuth,
)
from peek_core_user._private.server.auth_connectors.LdapAuth import LdapAuth
from peek_plugin_base.storage.DbConnection import DbSessionCreator

logger = logging.getLogger(__name__)


class AdminAuthController:
    def __init__(self, dbSessionCreator: DbSessionCreator):
        self._dbSessionCreator: DbSessionCreator = dbSessionCreator

    def shutdown(self):
        pass

    @inlineCallbacks
    def check(self, userName, password) -> List[str]:
        if not password:
            raise LoginFailed("Password is empty")

        lastException = None

        # TRY INTERNAL IF ITS ENABLED
        try:
            internalAuth = InternalAuth(self._dbSessionCreator)
            if (yield internalAuth.isInternalAuthEnabled()):
                groupNames, _ = yield internalAuth.checkPassAsync(
                    userName, password, InternalAuth.FOR_SERVICE_ADMIN
                )
                return groupNames

        except Exception as e:
            lastException = e

        # TRY LDAP IF ITS ENABLED
        try:
            ldapAuth = LdapAuth(self._dbSessionCreator)
            if (yield ldapAuth.isLdapAuthEnabled()):
                # TODO Make the client tell us if it's for office or field
                internalUser = yield ldapAuth.getInternalUser(
                    userName, raiseNotLoggedInException=False
                )
                userUuid = internalUser.userUuid if internalUser else None

                groups, ldapLoggedInUser = yield ldapAuth.checkPassAsync(
                    userName, password, LdapAuth.FOR_SERVICE_ADMIN, userUuid
                )

                return groups, ldapLoggedInUser

        except Exception as e:
            lastException = e

        if lastException:
            raise lastException

        raise Exception(
            "No authentication handlers are enabled, enable one in settings"
        )



"""add UserLoggedIn title

Peek Plugin Database Migration Script

Revision ID: bc55dc1de80a
Revises: 8fbb7b3fa4d8
Create Date: 2025-01-15 23:56:41.364881

"""

# revision identifiers, used by Alembic.
revision = "bc55dc1de80a"
down_revision = "8fbb7b3fa4d8"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        "UserLoggedIn",
        sa.Column("userTitle", sa.VARCHAR(), nullable=True),
        schema="core_user",
    )


def downgrade():

    op.drop_column("UserLoggedIn", "userTitle", schema="core_user")



from peek_core_user._private.server.admin_backend.InternalGroupTableHandler import (
    makeInternalGroupTableHandler,
)
from peek_core_user._private.server.admin_backend.LdapSettingsHandler import (
    makeLdapSettingeHandler,
)
from peek_core_user._private.server.admin_backend.OtpSettingsHandler import (
    makeOtpSettingsHandler,
)
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from .InternalUserTableHandler import makeInternalUserTableHandler
from .SettingPropertyHandler import makeSettingPropertyHandler


def makeAdminBackendHandlers(
    tupleObservable: TupleDataObservableHandler, dbSessionCreator
):
    yield makeInternalUserTableHandler(tupleObservable, dbSessionCreator)
    yield makeInternalGroupTableHandler(tupleObservable, dbSessionCreator)

    yield makeSettingPropertyHandler(tupleObservable, dbSessionCreator)

    yield makeLdapSettingeHandler(tupleObservable, dbSessionCreator)
    yield makeOtpSettingsHandler(tupleObservable, dbSessionCreator)



from typing import Callable

from abc import ABCMeta, abstractmethod
from reactivex.subject import Subject
from twisted.internet.defer import Deferred

from peek_core_user.tuples.login.UserLoginResponseTuple import (
    UserLoginResponseTuple,
)
from peek_core_user.tuples.login.UserLogoutResponseTuple import (
    UserLogoutResponseTuple,
)

UserPostLoginHookCallable = Callable[[UserLoginResponseTuple], Deferred]

UserPostLogoutHookCallable = Callable[[UserLogoutResponseTuple], Deferred]


class UserFieldHookApiABC(metaclass=ABCMeta):
    @abstractmethod
    def postLoginObservable(self) -> Subject:
        """Get the post login observable

        Subscribers will be called just after the user has authenticated.

        :return: Subject for post-login events
        """

    @abstractmethod
    def postLogoffObservable(self) -> Subject:
        """Get the post logoff observable

        Subscribers will be called just after the user has logged off.

        :return: Subject for post-logoff events
        """

    @abstractmethod
    def addPostLoginHook(self, callable: UserPostLoginHookCallable) -> None:
        """Add Post Login Hook

        :param callable: This callable will be called just after the user has
                         authenticated
        """

    @abstractmethod
    def removePostLoginHook(self, callable: UserPostLoginHookCallable) -> None:
        """Remove Post Login Hook

        :param callable: This callable to remove from the hooks
        """

    @abstractmethod
    def addPostLogoutHook(self, callable: UserPostLogoutHookCallable) -> None:
        """Add Post Logout Hook

        :param callable: This callable will be called just after the user has
                         logged out
        """

    @abstractmethod
    def removePostLogoutHook(
        self, callable: UserPostLogoutHookCallable
    ) -> None:
        """Remove Post Logout Hook

        :param callable: This callable to remove from the hooks
        """



from datetime import datetime

import logging

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField

logger = logging.getLogger(__name__)


@addTupleType
class UserLoggedInInfoTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserLoggedInInfoTuple"

    loggedInDateTime: datetime = TupleField()
    userName: str = TupleField()
    userTitle: str | None = TupleField()
    deviceToken: str = TupleField()
    vehicle: str = TupleField()
    isFieldLogin: bool = TupleField()
    loggedInWithGroups: list[str] = TupleField()



"""Removed string limits

Peek Plugin Database Migration Script

Revision ID: 3fd59734e444
Revises: 45c62f8acd87
Create Date: 2019-06-07 09:42:17.692382

"""

# revision identifiers, used by Alembic.
revision = "3fd59734e444"
down_revision = "45c62f8acd87"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        "idx_InternalUserPassword_userId",
        "InternalUserPassword",
        ["userId"],
        unique=True,
        schema="core_user",
    )
    # ### end Alembic commands ###

    op.alter_column(
        "InternalGroup",
        "groupName",
        type_=sa.String(),
        nullable=False,
        schema="core_user",
    )
    op.alter_column(
        "InternalGroup",
        "importHash",
        type_=sa.String(),
        nullable=True,
        schema="core_user",
    )

    op.alter_column(
        "InternalUserPassword",
        "password",
        type_=sa.String(),
        nullable=False,
        schema="core_user",
    )

    op.alter_column(
        "InternalUser",
        "userName",
        type_=sa.String(),
        nullable=False,
        schema="core_user",
    )
    op.alter_column(
        "InternalUser",
        "userTitle",
        type_=sa.String(),
        nullable=False,
        schema="core_user",
    )
    op.alter_column(
        "InternalUser",
        "userUuid",
        type_=sa.String(),
        nullable=False,
        schema="core_user",
    )
    op.alter_column(
        "InternalUser",
        "importHash",
        type_=sa.String(),
        nullable=True,
        schema="core_user",
    )
    op.alter_column(
        "InternalUser",
        "mobile",
        type_=sa.String(),
        nullable=True,
        schema="core_user",
    )
    op.alter_column(
        "InternalUser",
        "email",
        type_=sa.String(),
        nullable=True,
        schema="core_user",
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "idx_InternalUserPassword_userId",
        table_name="InternalUserPassword",
        schema="core_user",
    )
    # ### end Alembic commands ###



import logging

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField

logger = logging.getLogger(__name__)


@addTupleType
class ImportInternalGroupTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "ImportInternalGroupTuple"

    #: The unique name of this group
    groupName: str = TupleField()

    #: The nice name of this group
    groupTitle: str = TupleField()



from vortex.Tuple import addTupleType, TupleField
from vortex.TupleAction import TupleActionABC

from peek_core_user._private.PluginNames import userPluginTuplePrefix


@addTupleType
class UserLoginAuthMethodAction(TupleActionABC):
    __tupleType__ = userPluginTuplePrefix + "UserLoginAuthMethodAction"

    AUTH_FOR_ADMIN = 1
    AUTH_FOR_OFFICE = 2
    AUTH_FOR_FIELD = 3

    userName: str = TupleField()
    authForService: int = TupleField()



from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@addTupleType
class UserDetailTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserDetailTuple"

    #:  The username / userid of the user, EG C917
    userName: str = TupleField()

    #:  The title of the user, EG 'Chief Wiggum'
    userTitle: str = TupleField()

    #:  An external system user uuid, EG 715903a7ebc14fb0afb00d432676c51c
    userUuid: str = TupleField()

    #:  The mobile number, EG +61 419 123 456
    mobile: Optional[str] = TupleField()

    #:  The email address, EG guy@place.com
    email: Optional[str] = TupleField()

    #:  A list of group names that this user belongs to
    groupNames: List[str] = TupleField()

    #: A field for additional data
    data: Optional[Dict] = TupleField()

    #: user creation stream
    importSource: str = TupleField()

    #:  The last date that the user logged in
    lastLoginDate: Optional[datetime] = TupleField()

    #:  The device token of the last device the user logged in on.
    lastLoginDeviceToken: Optional[str] = TupleField()

    #: user authenticator target - a value from
    #  `peek_core_user.tuples.UserAuthTargetEnum`
    authenticationTarget: str = TupleField()

    @property
    def userKey(self):
        return self.userName.lower()



import hashlib
import logging
from datetime import datetime

import pytz
from sqlalchemy.orm.exc import NoResultFound
from twisted.cred.error import LoginFailed
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger

from peek_core_user._private.server.auth_connectors.AuthABC import AuthABC
from peek_core_user._private.server.auth_connectors.LdapAuth import LdapAuth
from peek_core_user._private.server.controller.PasswordUpdateController import (
    PasswordUpdateController,
)
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.storage.InternalUserGroupTuple import (
    InternalUserGroupTuple,
)
from peek_core_user._private.storage.Setting import ADMIN_LOGIN_GROUP
from peek_core_user._private.storage.Setting import (
    INTERNAL_AUTH_ENABLED_FOR_ADMIN,
)
from peek_core_user._private.storage.Setting import FIELD_LOGIN_GROUP
from peek_core_user._private.storage.Setting import FIELD_OTP_LOGIN_GROUP
from peek_core_user._private.storage.Setting import OFFICE_LOGIN_GROUP
from peek_core_user._private.storage.Setting import globalSetting
from peek_core_user.server.UserDbErrors import UserNotFoundException
from peek_core_user.server.UserDbErrors import UserPasswordNotSetException
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user.tuples.constants.UserAuthTargetEnum import (
    UserAuthTargetEnum,
)

logger = logging.getLogger(__name__)


class InternalAuth(AuthABC):
    @inlineCallbacks
    def checkPassAsync(self, userName, password, forService):
        try:
            authenticatedUserPassword = yield self.getInternalUserAndPassword(
                userName
            )
        except NoResultFound:
            authenticatedUserPassword = None

        # if user not found
        if (
            not authenticatedUserPassword
            or not authenticatedUserPassword.InternalUserTuple
        ):
            logger.debug(f"User '{userName}' not found")
            raise UserNotFoundException(userName)

        # if user found but user is created by LDAP
        if (
            authenticatedUserPassword.InternalUserTuple.authenticationTarget
            == UserAuthTargetEnum.LDAP
        ):
            # delegate to LDAPAuth
            return (
                yield LdapAuth(self._dbSessionCreator).checkPassAsync(
                    userName, password, forService
                )
            )

        isPasswordOtp = yield self.checkIfUserAbleToLoginWithOtp(userName)

        if (
            not isPasswordOtp
            and not authenticatedUserPassword.InternalUserPassword
        ):
            logger.debug(f"Password is not set for user {userName}")
            raise UserPasswordNotSetException(userName)

        # Check if the user is actually logged into this device.
        if isPasswordOtp:
            return (
                yield self._authenticateViaInternalOtp(
                    authenticatedUserPassword,
                    password,
                )
            )

        return (
            yield self._authenticateViaInternalPassword(
                authenticatedUserPassword, password, forService
            )
        )

    @inlineCallbacks
    def _authenticateViaInternalOtp(self, authenticatedUser, password):
        internalUserTuple = authenticatedUser[0]
        now = datetime.now(pytz.utc)
        attempted = hashlib.sha256(password.encode()).hexdigest()
        if not (
            internalUserTuple.oneTimePasscode == attempted
            and now <= internalUserTuple.oneTimePasscodeExpiry
        ):
            msg = (
                f"Peek InternalAuth: "
                f"username or one-time password is incorrect or expired"
                f"for user "
                f"'{internalUserTuple.userTitle}'"
                f"({internalUserTuple.userName})"
            )
            logger.debug(msg)
            raise LoginFailed(msg)

        groups = yield self._getUserGroupsByUserId(
            self._dbSessionCreator, internalUserTuple.id
        )

        groupNames = [g.groupName for g in groups]

        hasAceessForService = yield self.checkIfUserInUserGroup(
            self._dbSessionCreator, self.FOR_SERVICE_FIELD_OTP, groups
        )

        if not hasAceessForService:
            msg = (
                f"Peek InternalAuth: "
                f"User '{internalUserTuple.userTitle}'"
                f"({internalUserTuple.userName}) "
                f"is not granted permission "
                f"to service {self.getForServiceName(self.FOR_SERVICE_FIELD_OTP)}. "
                f"Please check membership of related groups, "
                f"current joined groups of the user are {', '.join(groupNames)}"
            )
            logger.debug(msg)
            raise LoginFailed(msg)

        return groupNames, authenticatedUser.InternalUserTuple

    @inlineCallbacks
    def _authenticateViaInternalPassword(
        self, authenticatedUser, password, forService
    ):
        internalUserTuple = authenticatedUser[0]
        internalUserPassword = authenticatedUser[1]
        if internalUserPassword.password != PasswordUpdateController.hashPass(
            password
        ):
            msg = (
                f"Username or password is incorrect "
                f"for user "
                f"'{internalUserTuple.userTitle}'"
                f"({internalUserTuple.userName})"
            )
            logger.debug(msg)
            raise LoginFailed(msg)

        groups = yield self._getUserGroupsByUserId(
            self._dbSessionCreator, internalUserPassword.userId
        )

        hasAceessForService = yield self.checkIfUserInUserGroup(
            self._dbSessionCreator, forService, groups
        )

        groupNames = [g.groupName for g in groups]

        if not hasAceessForService:
            msg = (
                f"Peek InternalAuth: "
                f"User '{internalUserTuple.userTitle}'"
                f"({internalUserTuple.userName}) "
                f"is not granted permission "
                f"to service {self.getForServiceName(forService)}. "
                f"Please check membership of related groups, "
                f"current joined groups of the user are {', '.join(groupNames)}"
            )
            logger.debug(msg)
            raise LoginFailed(msg)

        return groupNames, authenticatedUser.InternalUserTuple

    @inlineCallbacks
    def checkIfUserAbleToLoginWithOtp(self, userName: str) -> bool:
        authSettings = yield self.getAuthSettings()

        if not authSettings.internalFieldEnabled:
            return False
        else:
            groups = yield self._getUserGroupsByUserName(
                self._dbSessionCreator, userName
            )

            return (
                yield self.checkIfUserInUserGroup(
                    dbSessionCreator=self._dbSessionCreator,
                    forService=AuthABC.FOR_SERVICE_FIELD_OTP,
                    groups=groups,
                )
            )

    @deferToThreadWrapWithLogger(logger)
    def _getUserGroupsByUserId(self, dbSessionCreator, userId):
        ormSession = dbSessionCreator()
        try:
            return (
                ormSession.query(InternalGroupTuple)
                .join(InternalUserGroupTuple)
                .filter(InternalUserGroupTuple.userId == userId)
                .all()
            )
        finally:
            ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def _getUserGroupsByUserName(
        self, dbSessionCreator, userName
    ) -> Deferred[list[InternalGroupTuple]]:
        ormSession = dbSessionCreator()
        try:
            groups = (
                ormSession.query(InternalGroupTuple)
                .join(
                    InternalUserGroupTuple,
                    onclause=InternalUserGroupTuple.groupId
                    == InternalGroupTuple.id,
                )
                .join(
                    InternalUserTuple,
                    onclause=InternalUserGroupTuple.userId
                    == InternalUserTuple.id,
                )
                .filter(InternalUserTuple.userName == userName)
                .all()
            )
            return groups
        finally:
            ormSession.close()

    @classmethod
    @deferToThreadWrapWithLogger(logger)
    def checkIfUserInUserGroup(
        cls, dbSessionCreator, forService, groups: list[InternalGroupTuple]
    ) -> bool:
        ormSession = dbSessionCreator()
        try:
            groupNames = [g.groupName for g in groups]
            if forService == cls.FOR_SERVICE_ADMIN:
                adminGroup = globalSetting(ormSession, ADMIN_LOGIN_GROUP)
                if adminGroup not in set(groupNames):
                    return False

            elif forService == cls.FOR_SERVICE_OFFICE:
                officeGroup = globalSetting(ormSession, OFFICE_LOGIN_GROUP)
                if officeGroup not in set(groupNames):
                    return False

            elif (
                forService == cls.FOR_SERVICE_FIELD
                or forService == cls.FOR_SERVICE_FIELD_OTP
            ):
                fieldGroup = globalSetting(ormSession, FIELD_LOGIN_GROUP)
                if fieldGroup not in set(groupNames):
                    return False

                if forService == cls.FOR_SERVICE_FIELD_OTP:
                    fieldGroup = globalSetting(
                        ormSession, FIELD_OTP_LOGIN_GROUP
                    )
                    if fieldGroup not in set(groupNames):
                        return False

            else:
                msg = (
                    f"Peek InternalAuth: Unhandled forService type "
                    f"{forService}({cls.getForServiceName(forService)})"
                )
                logger.debug(msg)
                raise Exception(msg)
            return True
        finally:
            ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def isInternalAuthEnabled(self):
        ormSession = self._dbSessionCreator()
        try:
            return globalSetting(ormSession, INTERNAL_AUTH_ENABLED_FOR_ADMIN)
            # No commit needed, we only query

        finally:
            ormSession.close()



import hashlib
import json
import logging
import random
from datetime import datetime
from datetime import timedelta
from smtplib import SMTPException
from typing import List
from typing import NamedTuple
from typing import Tuple

import pytz
from peek_core_email.server.EmailApiABC import EmailApiABC
from twisted.cred.error import LoginFailed
from twisted.internet import reactor
from twisted.internet import task
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.server.api.UserFieldHookApi import UserFieldHookApi
from peek_core_user._private.server.api.UserInfoApi import UserInfoApi
from peek_core_user._private.server.auth_connectors.AuthABC import AuthABC
from peek_core_user._private.server.auth_connectors.CryptoUtil import (
    decryptAES256GCM,
)
from peek_core_user._private.server.auth_connectors.InternalAuth import (
    InternalAuth,
)
from peek_core_user._private.server.auth_connectors.LdapAuth import LdapAuth
from peek_core_user._private.server.util.OtpWordlist import OTP_WORDLIST
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.Setting import OTP_NUMBER_OF_OPTIONS
from peek_core_user._private.storage.Setting import OTP_NUMBER_OF_WORDS
from peek_core_user._private.storage.Setting import OTP_VALID_SECONDS
from peek_core_user._private.storage.Setting import oneTimePasscodeSetting
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_core_user._private.tuples.LoggedInUserStatusTuple import (
    LoggedInUserStatusTuple,
)
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from peek_core_user.server.UserDbErrors import (
    UserIsNotLoggedInToThisDeviceError,
)
from peek_core_user.tuples.login.UserLoginAction import UserLoginAction
from peek_core_user.tuples.login.UserLoginAuthMethodAction import (
    UserLoginAuthMethodAction,
)
from peek_core_user.tuples.login.UserLoginAuthMethodResponseTuple import (
    UserLoginAuthMethodResponseTuple,
)
from peek_core_user.tuples.login.UserLoginOtpAction import UserLoginOtpAction
from peek_core_user.tuples.login.UserLoginOtpResponseTuple import (
    UserLoginOtpResponseTuple,
)
from peek_core_user.tuples.login.UserLoginResponseTuple import (
    UserLoginResponseTuple,
)
from peek_core_user.tuples.login.UserLogoutAction import UserLogoutAction
from peek_core_user.tuples.login.UserLogoutResponseTuple import (
    UserLogoutResponseTuple,
)
from peek_plugin_base.storage.DbConnection import DbSessionCreator

logger = logging.getLogger(__name__)

USER_ALREADY_LOGGED_ON_KEY = "pl-user.USER_ALREADY_LOGGED_ON"
DEVICE_ALREADY_LOGGED_ON_KEY = "pl-user.DEVICE_ALREADY_LOGGED_ON_KEY"


class _ForceLogout:
    def __init__(self, userUuid, deviceToken):
        self._userUuid = userUuid
        self._deviceToken = deviceToken

    def forceDbLogout(self, ormSession):
        ormSession.query(UserLoggedIn).filter(
            UserLoggedIn.userUuid == self._userUuid
        ).filter(UserLoggedIn.deviceToken == self._deviceToken).delete(
            synchronize_session=False
        )

    def notify(self, clientTupleObservable):
        clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                UserLoggedInTuple.tupleType(),
                selector=dict(deviceToken=self._deviceToken),
            )
        )


class _GeneratedOtp(NamedTuple):
    otp: list[str]
    otpOptions: list[str]
    forMobile: str
    otpValidFrom: datetime
    otpValidTo: datetime


class LoginLogoutController:
    def __init__(
        self,
        deviceApi: DeviceApiABC,
        emailApi: EmailApiABC,
        dbSessionCreator: DbSessionCreator,
    ):
        self._deviceApi: DeviceApiABC = deviceApi
        self._emailApi: EmailApiABC = emailApi
        self._fieldServiceHookApi: UserFieldHookApi = None
        self._infoApi: UserInfoApi = None
        self._dbSessionCreator: DbSessionCreator = dbSessionCreator
        self._clientTupleObservable: TupleDataObservableHandler = None
        self._adminTupleObservable: TupleDataObservableHandler = None

    def setup(
        self,
        clientTupleObservable,
        adminTupleObservable,
        hookApi: UserFieldHookApi,
        infoApi: UserInfoApi,
    ):
        self._clientTupleObservable = clientTupleObservable
        self._adminTupleObservable = adminTupleObservable
        self._fieldServiceHookApi = hookApi
        self._infoApi = infoApi

    def shutdown(self):
        self._clientTupleObservable = None
        self._fieldServiceHookApi = None
        self._infoApi = None

    @inlineCallbacks
    def _checkPassAsync(
        self,
        userName,
        encryptedPassword,
        passphrase,
        isFieldService: bool,
        userUuid=None,
    ) -> Tuple[List[str], InternalUserTuple]:
        if not encryptedPassword or not passphrase:
            raise LoginFailed("Password is empty")

        password = decryptAES256GCM(
            encodedBase64String=encryptedPassword, passphrase=passphrase
        )

        # TODO Make the client tell us if it's for office or field

        lastException = None

        forService = AuthABC.FOR_SERVICE_OFFICE
        if isFieldService:
            forService = AuthABC.FOR_SERVICE_FIELD

        internalAuth = InternalAuth(self._dbSessionCreator)
        authSettings = yield internalAuth.getAuthSettings()

        try:
            # TRY INTERNAL IF ITS ENABLED
            if (
                forService == AuthABC.FOR_SERVICE_FIELD
                and authSettings.internalFieldEnabled
            ):
                return (
                    yield internalAuth.checkPassAsync(
                        userName, password, forService
                    )
                )

            if (
                forService == AuthABC.FOR_SERVICE_OFFICE
                and authSettings.internalOfficeEnabled
            ):
                return (
                    yield internalAuth.checkPassAsync(
                        userName, password, forService
                    )
                )

        except Exception as e:
            lastException = e

        # TRY LDAP IF ITS ENABLED
        try:
            ldapAuth = LdapAuth(self._dbSessionCreator)
            if authSettings.ldapAuthEnabled:
                return (
                    yield ldapAuth.checkPassAsync(
                        userName, password, forService, userUuid
                    )
                )

        except Exception as e:
            lastException = e

        if lastException:
            # ALWAYS pause for 2 seconds if there is a failure, it could be
            # an incorrect password attempt.
            # noinspection PyTypeChecker
            yield task.deferLater(reactor, 2, lambda: None)
            raise lastException

        raise Exception(
            "No authentication handlers are enabled, enable one in settings"
        )

    @deferToThreadWrapWithLogger(logger)
    def _logoutInDb(
        self, logoutTuple: UserLogoutAction, raiseNotLoggedInException=True
    ):
        """
        Returns Deferred[UserLogoutResponseTuple]
        """

        ormSession = self._dbSessionCreator()
        try:
            # Check if the user is actually logged into this device.
            qry = (
                ormSession.query(UserLoggedIn)
                .filter(UserLoggedIn.userName == logoutTuple.userName)
                .filter(UserLoggedIn.deviceToken == logoutTuple.deviceToken)
            )

            if qry.count() == 0:
                if raiseNotLoggedInException:
                    raise UserIsNotLoggedInToThisDeviceError(
                        logoutTuple.userName
                    )
                else:
                    return

            ormSession.delete(qry.one())
            ormSession.commit()

        finally:
            ormSession.close()

    @inlineCallbacks
    def logout(self, logoutTuple: UserLogoutAction) -> Deferred:
        """Logout

        :param logoutTuple: The tuple containing the information to process
                                for the logout.

        :return A deferred that fires with List[UserLogoutResponseTuple]
        """

        deviceDescription = yield self._deviceApi.deviceDescription(
            logoutTuple.deviceToken
        )

        response = UserLogoutResponseTuple(
            userName=logoutTuple.userName,
            deviceToken=logoutTuple.deviceToken,
            deviceDescription=deviceDescription,
            acceptedWarningKeys=logoutTuple.acceptedWarningKeys,
            succeeded=True,
        )

        if logoutTuple.isFieldService:
            # Give the hooks a chance to fail the logout
            yield self._fieldServiceHookApi.callLogoutHooks(response)

        # If there are no problems, proceed with the logout.
        try:
            if response.succeeded:
                yield self._logoutInDb(logoutTuple)

        finally:
            # Delay this, otherwise the user gets kicked off before getting
            # the nice success message
            reactor.callLater(
                0.05, self._sendLogoutUpdate, logoutTuple.deviceToken
            )

        self._adminTupleObservable.notifyOfTupleUpdateForTuple(
            LoggedInUserStatusTuple.tupleType()
        )

        return response

    def _sendLogoutUpdate(self, deviceToken: str):
        self._clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                UserLoggedInTuple.tupleType(),
                selector=dict(deviceToken=deviceToken),
            )
        )

    @inlineCallbacks
    def _loginInDb(self, loginTuple: UserLoginAction):
        blockMultipleLogins = loginTuple.isFieldService

        userDetail = yield self._infoApi.user(loginTuple.userName)

        # This will login from the internal user if one already exists or
        # login from LDAP and create a user if an internal does not exist
        groups, _ = yield self._checkPassAsync(
            userName=loginTuple.userName,
            encryptedPassword=loginTuple.password,
            passphrase=loginTuple.uuid,
            isFieldService=blockMultipleLogins,
            userUuid=userDetail.userUuid if userDetail else None,
        )
        return (yield self._loginInDbInThread(loginTuple, groups))

    @deferToThreadWrapWithLogger(logger)
    def _loginInDbInThread(
        self, loginTuple: UserLoginAction, groups: list[str]
    ):
        """
        Returns Deferred[UserLoginResponseTuple]

        """

        userName = loginTuple.userName
        acceptedWarningKeys = set(loginTuple.acceptedWarningKeys)
        deviceToken = loginTuple.deviceToken
        vehicle = loginTuple.vehicleId
        isFieldService = loginTuple.isFieldService
        blockMultipleLogins = isFieldService

        forceLogouter = None

        responseTuple = UserLoginResponseTuple(
            userName=userName,
            userToken="Not implemented",
            succeeded=False,
            acceptedWarningKeys=loginTuple.acceptedWarningKeys,
            vehicleId=loginTuple.vehicleId,
        )

        if not deviceToken:
            raise Exception("peekToken must be supplied")

        thisDeviceDescription = self._deviceApi.deviceDescriptionBlocking(
            deviceToken
        )

        userDetail = self._infoApi.userBlocking(userName)

        # check user group and user password
        ormSession = self._dbSessionCreator()
        try:
            if not userDetail:
                responseTuple.setFailed()
                return responseTuple

            userKey = userDetail.userKey
            userUuid = userDetail.userUuid
            loginTuple.userName = userDetail.userName
            responseTuple.userName = userDetail.userName
            responseTuple.userDetail = userDetail

            # Find any current login sessions
            userLoggedIn = (
                ormSession.query(UserLoggedIn)
                .filter(UserLoggedIn.userUuid == userUuid)
                .filter(UserLoggedIn.isFieldLogin == isFieldService)
                .all()
            )
            userLoggedIn = userLoggedIn[0] if userLoggedIn else None

            loggedInElsewhere = (
                ormSession.query(UserLoggedIn)
                .filter(UserLoggedIn.deviceToken != deviceToken)
                .filter(UserLoggedIn.userUuid == userUuid)
                .filter(UserLoggedIn.isFieldLogin == isFieldService)
                .all()
            )

            if blockMultipleLogins and len(loggedInElsewhere) not in (0, 1):
                raise Exception(
                    "Found more than 1 ClientDevice for"
                    + (" token %s" % deviceToken)
                )

            loggedInElsewhere = (
                loggedInElsewhere[0] if loggedInElsewhere else None
            )

            sameDevice = userLoggedIn and loggedInElsewhere is None

            # If the user is logged in, but not to this client device, raise exception
            if blockMultipleLogins and userLoggedIn and not sameDevice:
                if USER_ALREADY_LOGGED_ON_KEY in acceptedWarningKeys:
                    forceLogouter = _ForceLogout(
                        userUuid, loggedInElsewhere.deviceToken
                    )

                    forceLogouter.forceDbLogout(ormSession)

                    userLoggedIn = False

                else:
                    otherDeviceDescription = (
                        self._deviceApi.deviceDescriptionBlocking(
                            loggedInElsewhere.deviceToken
                        )
                    )

                    # This is false if the logged in device has been removed from
                    # enrollment
                    if otherDeviceDescription:
                        responseTuple.setFailed()
                        responseTuple.addWarning(
                            USER_ALREADY_LOGGED_ON_KEY,
                            "User %s is already logged in, on device %s"
                            % (userName, otherDeviceDescription),
                        )

                        return responseTuple

                    # Else, The old device has been deleted,
                    # Just let them login to the same device.

                    forceLogouter = _ForceLogout(
                        loggedInElsewhere.userUuid,
                        loggedInElsewhere.deviceToken,
                    )
                    forceLogouter.forceDbLogout(ormSession)

            # If we're logging into the same device, but already logged in
            if sameDevice:  # Logging into the same device
                sameDeviceDescription = (
                    self._deviceApi.deviceDescriptionBlocking(
                        userLoggedIn.deviceToken
                    )
                )

                responseTuple.deviceToken = userLoggedIn.deviceToken
                responseTuple.deviceDescription = sameDeviceDescription
                responseTuple.succeeded = True
                return responseTuple

            anotherUserOnThatDevice = (
                ormSession.query(UserLoggedIn)
                .filter(UserLoggedIn.deviceToken == deviceToken)
                .filter(UserLoggedIn.userUuid != userUuid)
                .all()
            )

            if anotherUserOnThatDevice:
                anotherUserOnThatDevice = anotherUserOnThatDevice[0]
                if DEVICE_ALREADY_LOGGED_ON_KEY in acceptedWarningKeys:
                    forceLogouter = _ForceLogout(
                        anotherUserOnThatDevice.userUuid,
                        anotherUserOnThatDevice.deviceToken,
                    )
                    forceLogouter.forceDbLogout(ormSession)

                else:
                    responseTuple.setFailed()
                    responseTuple.addWarning(
                        DEVICE_ALREADY_LOGGED_ON_KEY,
                        "User %s is currently logged into this device : %s"
                        % (
                            anotherUserOnThatDevice.userName,
                            thisDeviceDescription,
                        ),
                    )

                    return responseTuple

            # Create the user logged in entry
            newUser = UserLoggedIn(
                userTitle=userDetail.userTitle,
                userName=userName.lower(),
                userKey=userKey.lower(),
                userUuid=userUuid,
                loggedInDateTime=datetime.now(pytz.utc),
                deviceToken=deviceToken,
                vehicle=vehicle,
                isFieldLogin=isFieldService,
                loggedInWithGroups=json.dumps(groups if groups else []),
            )
            ormSession.add(newUser)

            # Update the last login details
            updatedInternalUser = (
                ormSession.query(InternalUserTuple)
                .filter(InternalUserTuple.userUuid == userUuid)
                .one()
            )
            updatedInternalUser.lastLoginDate = datetime.now(pytz.utc)
            updatedInternalUser.lastLoginDeviceToken = deviceToken

            ormSession.commit()

            # Respond with a successful login
            responseTuple.deviceToken = deviceToken
            responseTuple.deviceDescription = thisDeviceDescription
            responseTuple.succeeded = True
            return responseTuple

        finally:
            ormSession.close()

            if forceLogouter:
                forceLogouter.notify(self._clientTupleObservable)

    @inlineCallbacks
    def login(self, loginTuple: UserLoginAction):
        """
        Returns Deferred[UserLoginResponseTuple]

        """
        loginResponse = None
        try:
            loginResponse = yield self._loginInDb(loginTuple)

            if loginTuple.isFieldService:
                yield self._fieldServiceHookApi.callLoginHooks(loginResponse)

        # except UserAlreadyLoggedInError as e:
        #     pass
        #
        # except DeviceAlreadyLoggedInError as e:
        #     pass
        #
        # except UserIsNotLoggedInToThisDeviceError as e:
        #     pass

        except Exception as e:
            # Log the user out again if the hooks fail
            logoutTuple = UserLogoutAction(
                userName=loginTuple.userName, deviceToken=loginTuple.deviceToken
            )

            # Force logout, we don't care if it works or not.
            try:
                yield self._logoutInDb(
                    logoutTuple, raiseNotLoggedInException=False
                )
            except UserIsNotLoggedInToThisDeviceError:
                pass

            logger.debug(f"User login failed: {e}")
            raise e

        self._clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                UserLoggedInTuple.tupleType(),
                selector=dict(deviceToken=loginTuple.deviceToken),
            )
        )

        self._adminTupleObservable.notifyOfTupleUpdateForTuple(
            LoggedInUserStatusTuple.tupleType()
        )

        return loginResponse

    def _forceLogout(self, ormSession, userUuid, deviceToken):
        ormSession.query(UserLoggedIn).filter(
            UserLoggedIn.userUuid == userUuid
        ).filter(UserLoggedIn.deviceToken == deviceToken).delete(
            synchronize_session=False
        )

        self._clientTupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                UserLoggedInTuple.tupleType(),
                selector=dict(deviceToken=deviceToken),
            )
        )

    @inlineCallbacks
    def queryUserLoginAuthMethod(
        self, userLoginAuthMethodAction: UserLoginAuthMethodAction
    ) -> Deferred[UserLoginAuthMethodResponseTuple]:
        responseTuple = UserLoginAuthMethodResponseTuple()
        responseTuple.userName = userLoginAuthMethodAction.userName

        if not userLoginAuthMethodAction.AUTH_FOR_FIELD:
            responseTuple.authMethod = (
                UserLoginAuthMethodResponseTuple.AUTH_METHOD_PASSWORD
            )
        else:
            internalAuth = InternalAuth(self._dbSessionCreator)
            isOtpAvailableForUser = (
                yield internalAuth.checkIfUserAbleToLoginWithOtp(
                    userName=userLoginAuthMethodAction.userName
                )
            )

            if isOtpAvailableForUser:
                responseTuple.authMethod = (
                    UserLoginAuthMethodResponseTuple.AUTH_METHOD_OTP
                )
            else:
                responseTuple.authMethod = (
                    UserLoginAuthMethodResponseTuple.AUTH_METHOD_PASSWORD
                )

        return responseTuple

    @deferToThreadWrapWithLogger(logger)
    def _generateOtpForUser(self, userName: str) -> _GeneratedOtp:
        startTime = datetime.now(pytz.utc)

        ormSession = self._dbSessionCreator()
        try:
            settings = oneTimePasscodeSetting(ormSession)

            validDuration = timedelta(seconds=settings[OTP_VALID_SECONDS])  # z
            otpOptions = list(
                sorted(
                    random.sample(OTP_WORDLIST, settings[OTP_NUMBER_OF_OPTIONS])
                )
            )  # y
            otp = list(
                random.sample(otpOptions, settings[OTP_NUMBER_OF_WORDS])
            )  # x
            _otpCode = "_".join(otp)
            otpCodeHashed = hashlib.sha256(_otpCode.encode()).hexdigest()

            user = (
                ormSession.query(InternalUserTuple)
                .filter(InternalUserTuple.userName == userName)
                .one()
            )

            user.oneTimePasscode = otpCodeHashed
            user.oneTimePasscodeExpiry = datetime.now(pytz.utc) + validDuration

            ormSession.commit()

            return _GeneratedOtp(
                otp=otp,
                otpOptions=otpOptions,
                forMobile=str(user.mobile),
                otpValidTo=startTime + validDuration,
                otpValidFrom=startTime,
            )

        finally:
            ormSession.close()

    @inlineCallbacks
    def _sendOtpCode(self, generatedOtp: _GeneratedOtp):
        otpStr = " ".join(generatedOtp.otp)
        mobile = generatedOtp.forMobile
        smsMessage = (
            f'Your one-time password is "{otpStr}". '
            f"Do not share this with anyone. "
            f"We will never contact you to request this code."
        )
        yield self._emailApi.sendSms(mobile, contents=smsMessage)

    @inlineCallbacks
    def loginRequestOtp(self, userLoginOtpAction: UserLoginOtpAction):
        responseTuple = UserLoginOtpResponseTuple()
        responseTuple.userName = userLoginOtpAction.userName

        internalAuth = InternalAuth(self._dbSessionCreator)
        isOtpAvailableForUser = (
            yield internalAuth.checkIfUserAbleToLoginWithOtp(
                userName=userLoginOtpAction.userName
            )
        )
        if not isOtpAvailableForUser:
            responseTuple.otpRequestStatus = (
                UserLoginOtpResponseTuple.STATUS_OTP_REQUEST_REJECTED
            )
        else:
            responseTuple.otpRequestStatus = (
                UserLoginOtpResponseTuple.STATUS_OTP_REQUEST_ACCEPTED
            )
            generatedOtp = yield self._generateOtpForUser(
                userName=userLoginOtpAction.userName
            )
            responseTuple.otpOptions = generatedOtp.otpOptions
            responseTuple.otpValidFrom = generatedOtp.otpValidFrom
            responseTuple.otpValidTo = generatedOtp.otpValidTo

            try:
                yield self._sendOtpCode(generatedOtp)

            except Exception as e:
                raise SMTPException(
                    "failed to send SMS, "
                    "please contact Peek Admin for email settings."
                )
        return responseTuple



"""separated password

Peek Plugin Database Migration Script

Revision ID: 2c6728ebc562
Revises: 9b65be31926e
Create Date: 2018-01-28 16:39:17.645627

"""

# revision identifiers, used by Alembic.
revision = "2c6728ebc562"
down_revision = "9b65be31926e"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.create_table(
        "InternalUserPassword",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("userId", sa.Integer(), nullable=False),
        sa.Column("password", sa.String(length=50), nullable=False),
        sa.ForeignKeyConstraint(
            ["userId"], ["core_user.InternalUser.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="core_user",
    )

    op.execute(
        """
    INSERT INTO core_user."InternalUserPassword" ("userId", "password")
    SELECT "id", "password"
    FROM core_user."InternalUser"
    WHERE "password" is not null
    """
    )

    op.drop_column("InternalUser", "password", schema="core_user")


def downgrade():
    raise NotImplementedError("Downgrade not implemented")



import base64
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend


def decryptAES256GCM(encodedBase64String: str, passphrase: str) -> str:
    concatBinaryString = base64.b64decode(encodedBase64String)

    # salt [16bytes] |
    # iv [12bytes] |
    # ciphertext (ciphertext [variable length] | tag [16bytes] )
    receivedSalt = concatBinaryString[0:16]
    receivedIv = concatBinaryString[16 : 16 + 12]
    receivedCiphertext = concatBinaryString[16 + 12 : -16]
    receivedTag = concatBinaryString[-16:]
    # Derive the key from the received salt and password
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=receivedSalt,
        iterations=100000,
        backend=default_backend(),
    )
    key = kdf.derive(passphrase.encode("utf-8"))
    # Decrypt the ciphertext using AES-GCM mode
    backend = default_backend()
    cipher = Cipher(
        algorithms.AES256(key), modes.GCM(receivedIv), backend=backend
    )
    decryptor = cipher.decryptor()
    decryptedData = decryptor.update(
        receivedCiphertext
    ) + decryptor.finalize_with_tag(receivedTag)

    return decryptedData.decode("utf-8")



import logging
from typing import List
from typing import Optional
from typing import Tuple

from vortex.rpc.RPC import vortexRPC

from peek_core_user._private.PluginNames import userPluginFilt
from peek_core_user._private.ldap_auth.ldap_auth import checkLdapAuth
from peek_core_user._private.storage.LdapSetting import LdapSetting
from peek_core_user._private.tuples.LdapLoggedInUserTuple import (
    LdapLoggedInUserTuple,
)
from peek_plugin_base.PeekVortexUtil import peekAgentName

logger = logging.getLogger(__name__)


class RpcForLogic:
    def __init__(self):
        pass

    def makeHandlers(self):
        """Make Handlers

        In this method we start all the RPC handlers
        start() returns an instance of itself so we can simply yield the result
        of the start method.

        """
        yield self.tryLdapLoginOnAgent.start(funcSelf=self)
        logger.debug("LogicService RPCs started")

    # -------------
    @vortexRPC(peekAgentName, additionalFilt=userPluginFilt)
    def tryLdapLoginOnAgent(
        self,
        username: str,
        password: str,
        ldapSetting: LdapSetting,
        userUuid: Optional[str],
    ) -> Tuple[List[str], LdapLoggedInUserTuple]:
        return checkLdapAuth(username, password, ldapSetting, userUuid)



import json
import logging

from peek_core_user._private.server.api.UserApi import UserApi
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_core_user._private.tuples.UserLoggedInTuple import UserLoggedInTuple
from peek_core_user.tuples.UserListItemTuple import UserListItemTuple
from sqlalchemy.orm.exc import NoResultFound
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class UserLoggedInTupleProvider(TuplesProviderABC):
    def __init__(self, dbSessionCreator, ourApi: UserApi):
        self._dbSessionCreator = dbSessionCreator
        self._ourApi = ourApi

        from peek_core_user.server.UserApiABC import UserApiABC

        assert isinstance(
            self._ourApi, UserApiABC
        ), "We didn't get a UserApiABC"

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Deferred:
        deviceToken = tupleSelector.selector["deviceToken"]

        session = self._dbSessionCreator()
        try:
            userLoggedIn = (
                session.query(UserLoggedIn)
                .filter(UserLoggedIn.deviceToken == deviceToken)
                .one()
            )

            internalUserTuple = (
                session.query(InternalUserTuple)
                .filter(InternalUserTuple.userUuid == userLoggedIn.userUuid)
                .one()
            )

            if userLoggedIn.loggedInWithGroups:
                loggedInWithGroups = json.loads(userLoggedIn.loggedInWithGroups)
            else:
                loggedInWithGroups = []

            userDetails = UserListItemTuple(
                userId=internalUserTuple.userName,
                displayName=internalUserTuple.userTitle,
                mobile=internalUserTuple.mobile,
            )

        except NoResultFound:
            userDetails = None
            loggedInWithGroups = []

        finally:
            session.close()

        tuples = [
            UserLoggedInTuple(
                deviceToken=deviceToken,
                userDetails=userDetails,
                userGroups=loggedInWithGroups,
            )
        ]

        payloadEnvelope = Payload(
            filt=filt, tuples=tuples
        ).makePayloadEnvelope()
        vortexMsg = payloadEnvelope.toVortexMsg()
        return vortexMsg



import logging
from collections import namedtuple
from typing import List

from twisted.cred.error import LoginFailed
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.VortexFactory import NoVortexException
from vortex.VortexFactory import VortexFactory

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.agent.RpcForLogic import RpcForLogic
from peek_core_user._private.ldap_auth.ldap_auth import checkLdapAuth
from peek_core_user._private.server.auth_connectors.AuthABC import AuthABC
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.LdapSetting import LdapSetting
from peek_core_user._private.storage.Setting import LDAP_AUTH_ENABLED
from peek_core_user._private.storage.Setting import LDAP_ENABLE_DOMAIN_SUPPORT
from peek_core_user._private.storage.Setting import LDAP_VERIFY_SSL
from peek_core_user._private.storage.Setting import globalSetting
from peek_platform.file_config.PeekFileConfigABC import PEEK_AGENT_SERVICE

__author__ = "synerty"

from peek_core_user._private.tuples.LdapLoggedInUserTuple import (
    LdapLoggedInUserTuple,
)
from peek_core_user.tuples.constants.UserAuthTargetEnum import (
    UserAuthTargetEnum,
)

logger = logging.getLogger(__name__)

import ldap


class LdapNotEnabledError(Exception):
    pass


_LdapAuthSettings = namedtuple(
    "_LdapAuthSettings", ["ldapSettings", "ldapVerifySsl", "ldapEnableDomain"]
)


class LdapAuth(AuthABC):
    @inlineCallbacks
    def checkPassAsync(self, userName, password, forService, userUuid=None):
        """Login User

        :param forService:
        :param userName: The username of the user.
        :param password: The users secret password.
        :param userUuid: The decoded objectSid of the user from LDAP
        :rtype
        """

        assert forService in (
            AuthABC.FOR_SERVICE_FIELD,
            AuthABC.FOR_SERVICE_OFFICE,
            AuthABC.FOR_SERVICE_ADMIN,
        ), "Unhandled authentication for service type"

        authSettings: _LdapAuthSettings = yield self._getSettings()

        if not authSettings.ldapEnableDomain and "@" in userName:
            raise Exception(
                "Please login with a username, not an email address"
            )

        if not authSettings.ldapVerifySsl:
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

        if not authSettings.ldapSettings:
            raise Exception("LDAPAuth: No LDAP servers configured.")

        firstException = None

        for ldapSetting in authSettings.ldapSettings:
            if forService == self.FOR_SERVICE_ADMIN:
                if not ldapSetting.adminEnabled:
                    continue

            elif forService == self.FOR_SERVICE_OFFICE:
                if not ldapSetting.desktopEnabled:
                    continue

            elif forService == self.FOR_SERVICE_FIELD:
                if not ldapSetting.mobileEnabled:
                    continue

            else:
                raise Exception(
                    "LDAPAuth: Unhandled forService type %s" % forService
                )

            if (
                "@" in userName
                and userName.split("@")[1] != ldapSetting.ldapDomain
            ):
                continue

            try:
                logger.info("Trying LDAP login for %s", ldapSetting.ldapDomain)

                # None, zero length string and just spaces.
                if (
                    not ldapSetting.agentHost
                    or not ldapSetting.agentHost.strip()
                ):
                    logger.debug("Trying LDAP on Logic Service")
                    (groups, ldapLoggedInUser) = checkLdapAuth(
                        userName, password, ldapSetting, userUuid
                    )

                else:
                    logger.debug(
                        "Trying LDAP on Agent %s", ldapSetting.agentHost
                    )
                    (
                        groups,
                        ldapLoggedInUser,
                    ) = yield self._forwardLdapAuthToAgent(
                        userName, password, ldapSetting, userUuid
                    )

                internalUser = yield self._getOrCreateInternalUserAsync(
                    ldapLoggedInUser
                )
                return groups, internalUser

            except LoginFailed as e:
                if not firstException:
                    firstException = e

        logger.error("Login failed for %s, %s", userName, str(firstException))

        if firstException:
            raise firstException

        raise LoginFailed("LDAPAuth: No LDAP providers found for this service")

    @inlineCallbacks
    def _forwardLdapAuthToAgent(
        self, userName, password, ldapSetting, userUuid
    ):
        agentHost = ldapSetting.agentHost
        try:
            vortexUuid = VortexFactory.getRemoteVortexInfoByIp(
                agentHost, PEEK_AGENT_SERVICE
            )

        except KeyError:
            logger.error(f"Could not find Peek Agent with hostname {agentHost}")
            raise LoginFailed(
                "Failed to login, ask an administrator to check " "the logs"
            )

        except NoVortexException as e:
            logger.error(str(e))
            raise LoginFailed(
                "Failed to login, ask an administrator to check " "the logs"
            )

        try:
            return (
                yield RpcForLogic.tryLdapLoginOnAgent.callForVortexUuid(
                    vortexUuid, userName, password, ldapSetting, userUuid
                )
            )
        except Exception as e:
            logger.debug("Login failed on agent, here is the error: %s", e)
            raise LoginFailed(
                str(e)
                .splitlines()[-1]
                .replace("twisted.cred.error.LoginFailed: LDAPAuth: ", "")
            )

    @deferToThreadWrapWithLogger(logger)
    def _getOrCreateInternalUserAsync(
        self, ldapLoggedInUser: LdapLoggedInUserTuple
    ) -> List[LdapSetting]:
        session = self._dbSessionCreator()
        try:
            internalUser = self._getOrCreateInternalUserBlocking(
                session, ldapLoggedInUser
            )
            session.expunge_all()
            session.commit()

            return internalUser

        finally:
            session.close()

    # noinspection PyMethodMayBeStatic
    def _getOrCreateInternalUserBlocking(
        self, dbSession, ldapLoggedInUser: LdapLoggedInUserTuple
    ) -> InternalUserTuple:
        internalUser = (
            dbSession.query(InternalUserTuple)
            .filter(InternalUserTuple.userUuid == ldapLoggedInUser.userUuid)
            .first()
        )

        # do no create, return the existing user
        if internalUser:
            logger.info("Found existing internal user %s", internalUser.userKey)
            if "@" not in internalUser.userKey:
                internalUser.userKey = (
                    internalUser.userName
                    if "@" in internalUser.userName
                    else (
                        "%s@%s"
                        % (internalUser.userName, ldapLoggedInUser.ldapDomain)
                    )
                )
                dbSession.merge(internalUser)
                dbSession.commit()

            return internalUser

        userKey = "%s@%s" % (
            ldapLoggedInUser.username,
            ldapLoggedInUser.ldapDomain,
        )
        logger.info("Creating new internal user: %s", userKey)
        newInternalUser = InternalUserTuple(
            userName=ldapLoggedInUser.username.lower(),
            userKey=(
                ldapLoggedInUser.username
                if "@" in ldapLoggedInUser.username
                else userKey
            ).lower(),
            userTitle="%s (%s)"
            % (ldapLoggedInUser.userTitle, ldapLoggedInUser.ldapName),
            userUuid=ldapLoggedInUser.userUuid,
            email=(
                ldapLoggedInUser.email.lower()
                if ldapLoggedInUser.email
                else None
            ),
            authenticationTarget=UserAuthTargetEnum.LDAP,
            importSource="LDAP",
            # importHash e.g. 'peek_core_user.LDAPAuth:<md5 hash>'
            importHash=f"{userPluginTuplePrefix}LDAPAuth:{ldapLoggedInUser.userUuid}",
        )

        try:
            dbSession.add(newInternalUser)
        except Exception as e:
            logger.info(e)
            raise LoginFailed(
                "Failed to create Internal User. Use the full name <username>@<ldap-domain> to login"
            )
        dbSession.commit()
        return newInternalUser

    @deferToThreadWrapWithLogger(logger)
    def _getSettings(self) -> _LdapAuthSettings:
        session = self._dbSessionCreator()
        try:
            # Check if the user is actually logged into this device.
            ldapSettings = session.query(LdapSetting).all()
            ldapVerifySsl = globalSetting(session, LDAP_VERIFY_SSL)
            ldapEnableDomain = globalSetting(
                session, LDAP_ENABLE_DOMAIN_SUPPORT
            )

            session.expunge_all()

            return _LdapAuthSettings(
                ldapSettings=ldapSettings,
                ldapVerifySsl=ldapVerifySsl,
                ldapEnableDomain=ldapEnableDomain,
            )

            # No commit needed, we only query

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def isLdapAuthEnabled(self) -> _LdapAuthSettings:
        session = self._dbSessionCreator()
        try:
            return globalSetting(session, LDAP_AUTH_ENABLED)
            # No commit needed, we only query

        finally:
            session.close()
